@@ -13,6 +13,34 @@ import sys
 
 class ApplyPatchTestCase(unittest.TestCase):
 
+    def test_js_file(self):
+        with open('./tests.js', 'r') as f:
+            tests = json.load(f)
+            for test in tests:
+                try:
+                    if 'expected' not in test:
+                        continue
+                    result = jsonpatch.apply_patch(test['doc'], test['patch'])
+                    self.assertEqual(result, test['expected'])
+                except Exception:
+                    if test.get('error'):
+                        continue
+                    else:
+                        raise
+
+    def test_success_if_replaced_dict(self):
+        src = [{'a': 1}, {'b': 2}]
+        dst = [{'a': 1, 'b': 2}]
+        patch = jsonpatch.make_patch(src, dst)
+        self.assertEqual(patch.apply(src), dst)
+
+    def test_success_if_raise_no_error(self):
+        src = [{}]
+        dst = [{'key': ''}]
+        patch = jsonpatch.make_patch(src, dst)
+        patch.apply(src)
+        self.assertTrue(True)
+
     def test_apply_patch_from_string(self):
         obj = {'foo': 'bar'}
         patch = '[{"op": "add", "path": "/baz", "value": "qux"}]'
@@ -239,7 +267,6 @@ class EqualityTestCase(unittest.TestCase):
         self.assertEqual(json.dumps(patch_obj), patch.to_string())
 
 
-
 class MakePatchTestCase(unittest.TestCase):
 
     def test_apply_patch_to_copy(self):
@@ -299,7 +326,9 @@ class MakePatchTestCase(unittest.TestCase):
                    }
         self.assertEqual(expected, res)
 
-    def test_should_just_add_new_item_not_rebuild_all_list(self):
+    # TODO: this test is currently disabled, as the optimized patch is
+    # not ideal
+    def _test_should_just_add_new_item_not_rebuild_all_list(self):
         src = {'foo': [1, 2, 3]}
         dst = {'foo': [3, 1, 2, 3]}
         patch = list(jsonpatch.make_patch(src, dst))
@@ -308,33 +337,6 @@ class MakePatchTestCase(unittest.TestCase):
         res = jsonpatch.apply_patch(src, patch)
         self.assertEqual(res, dst)
 
-    def test_use_replace_instead_of_remove_add(self):
-        src = {'foo': [1, 2, 3]}
-        dst = {'foo': [3, 2, 3]}
-        patch = list(jsonpatch.make_patch(src, dst))
-        self.assertEqual(len(patch), 1)
-        self.assertEqual(patch[0]['op'], 'replace')
-        res = jsonpatch.apply_patch(src, patch)
-        self.assertEqual(res, dst)
-
-    def test_use_move_instead_of_remove_add(self):
-        src = {'foo': [4, 1, 2, 3]}
-        dst = {'foo': [1, 2, 3, 4]}
-        patch = list(jsonpatch.make_patch(src, dst))
-        self.assertEqual(len(patch), 1)
-        self.assertEqual(patch[0]['op'], 'move')
-        res = jsonpatch.apply_patch(src, patch)
-        self.assertEqual(res, dst)
-
-    def test_use_move_instead_of_add_remove(self):
-        src = {'foo': [1, 2, 3]}
-        dst = {'foo': [3, 1, 2]}
-        patch = list(jsonpatch.make_patch(src, dst))
-        self.assertEqual(len(patch), 1)
-        self.assertEqual(patch[0]['op'], 'move')
-        res = jsonpatch.apply_patch(src, patch)
-        self.assertEqual(res, dst)
-
     def test_escape(self):
         src = {"x/y": 1}
         dst = {"x/y": 2}
@@ -359,6 +361,198 @@ class MakePatchTestCase(unittest.TestCase):
         res = patch.apply(src)
         self.assertEqual(res, dst)
 
+    def test_issue40(self):
+        """ Tests an issue in _split_by_common_seq reported in #40 """
+
+        src = [8, 7, 2, 1, 0, 9, 4, 3, 5, 6]
+        dest = [7, 2, 1, 0, 9, 4, 3, 6, 5, 8]
+        jsonpatch.make_patch(src, dest)
+
+    def test_issue76(self):
+        """ Make sure op:remove does not include a 'value' field """
+
+        src = { "name": "fred", "friend": "barney", "spouse": "wilma" }
+        dst = { "name": "fred", "spouse": "wilma" }
+        expected = [{"path": "/friend", "op": "remove"}]
+        patch = jsonpatch.make_patch(src, dst)
+        self.assertEqual(patch.patch, expected)
+        res = jsonpatch.apply_patch(src, patch)
+        self.assertEqual(res, dst)
+
+    def test_json_patch(self):
+        old = {
+            'queue': {'teams_out': [{'id': 3, 'reason': 'If tied'}, {'id': 5, 'reason': 'If tied'}]},
+        }
+        new = {
+            'queue': {'teams_out': [{'id': 5, 'reason': 'If lose'}]}
+        }
+        patch = jsonpatch.make_patch(old, new)
+        new_from_patch = jsonpatch.apply_patch(old, patch)
+        self.assertEqual(new, new_from_patch)
+
+    def test_arrays_one_element_sequences(self):
+        """ Tests the case of multiple common one element sequences inside an array """
+        # see https://github.com/stefankoegl/python-json-patch/issues/30#issuecomment-155070128
+        src = [1,2,3]
+        dst = [3,1,4,2]
+        patch = jsonpatch.make_patch(src, dst)
+        res = jsonpatch.apply_patch(src, patch)
+        self.assertEqual(res, dst)
+
+    def test_list_in_dict(self):
+        """ Test patch creation with a list within a dict, as reported in #74
+
+        https://github.com/stefankoegl/python-json-patch/issues/74 """
+        old = {'key': [{'someNumber': 0, 'someArray': [1, 2, 3]}]}
+        new = {'key': [{'someNumber': 0, 'someArray': [1, 2, 3, 4]}]}
+        patch = jsonpatch.make_patch(old, new)
+        new_from_patch = jsonpatch.apply_patch(old, patch)
+        self.assertEqual(new, new_from_patch)
+
+    def test_nested(self):
+        """ Patch creation with nested dicts, as reported in #41
+
+        https://github.com/stefankoegl/python-json-patch/issues/41 """
+        old = {'school':{'names':['Kevin','Carl']}}
+        new = {'school':{'names':['Carl','Kate','Kevin','Jake']}}
+        patch = jsonpatch.JsonPatch.from_diff(old, new)
+        new_from_patch = jsonpatch.apply_patch(old, patch)
+        self.assertEqual(new, new_from_patch)
+
+
+class OptimizationTests(unittest.TestCase):
+    def test_use_replace_instead_of_remove_add(self):
+        src = {'foo': [1, 2, 3]}
+        dst = {'foo': [3, 2, 3]}
+        patch = list(jsonpatch.make_patch(src, dst))
+        self.assertEqual(len(patch), 1)
+        self.assertEqual(patch[0]['op'], 'replace')
+        res = jsonpatch.apply_patch(src, patch)
+        self.assertEqual(res, dst)
+
+    def test_use_replace_instead_of_remove_add_nested(self):
+        src = {'foo': [{'bar': 1, 'baz': 2}, {'bar': 2, 'baz': 3}]}
+        dst = {'foo': [{'bar': 1}, {'bar': 2, 'baz': 3}]}
+        patch = list(jsonpatch.make_patch(src, dst))
+
+        exp = [{'op': 'remove', 'path': '/foo/0/baz'}]
+        self.assertEqual(patch, exp)
+
+        res = jsonpatch.apply_patch(src, patch)
+        self.assertEqual(res, dst)
+
+    def test_use_move_instead_of_remove_add(self):
+        src = {'foo': [4, 1, 2, 3]}
+        dst = {'foo': [1, 2, 3, 4]}
+        patch = list(jsonpatch.make_patch(src, dst))
+        self.assertEqual(len(patch), 1)
+        self.assertEqual(patch[0]['op'], 'move')
+        res = jsonpatch.apply_patch(src, patch)
+        self.assertEqual(res, dst)
+
+    def test_use_move_instead_of_add_remove(self):
+        def fn(_src, _dst):
+            patch = list(jsonpatch.make_patch(_src, _dst))
+            # Check if there are only 'move' operations
+            for p in patch:
+                self.assertEqual(p['op'], 'move')
+            res = jsonpatch.apply_patch(_src, patch)
+            self.assertEqual(res, _dst)
+
+        fn({'foo': [1, 2, 3]}, {'foo': [3, 1, 2]})
+        fn([1, 2, 3], [3, 1, 2])
+        fn({'foo': [1, 2, 3]}, {'foo': [3, 2, 1]})
+        fn([1, 2, 3], [3, 2, 1])
+
+    def test_success_if_replace_inside_dict(self):
+        src = [{'a': 1, 'foo': {'b': 2, 'd': 5}}]
+        dst = [{'a': 1, 'foo': {'b': 3, 'd': 6}}]
+        patch = jsonpatch.make_patch(src, dst)
+        self.assertEqual(patch.apply(src), dst)
+
+    def test_success_if_replace_single_value(self):
+        src = [{'a': 1, 'b': 2, 'd': 5}]
+        dst = [{'a': 1, 'c': 3, 'd': 5}]
+        patch = jsonpatch.make_patch(src, dst)
+        self.assertEqual(patch.apply(src), dst)
+
+    def test_success_if_replaced_by_object(self):
+        src = [{'a': 1, 'b': 2, 'd': 5}]
+        dst = [{'d': 6}]
+        patch = jsonpatch.make_patch(src, dst)
+        self.assertEqual(patch.apply(src), dst)
+
+    def test_success_if_correct_patch_appied(self):
+        src = [{'a': 1}, {'b': 2}]
+        dst = [{'a': 1, 'b': 2}]
+        patch = jsonpatch.make_patch(src, dst)
+        self.assertEqual(patch.apply(src), dst)
+
+    def test_success_if_correct_expected_patch_appied(self):
+        src = [{"a": 1, "b": 2}]
+        dst = [{"b": 2, "c": 2}]
+        exp = [
+            {'path': '/0/a', 'op': 'remove'},
+            {'path': '/0/c', 'op': 'add', 'value': 2}
+        ]
+        patch = jsonpatch.make_patch(src, dst)
+        self.assertEqual(patch.patch, exp)
+        # verify that this patch does what we expect
+        res = jsonpatch.apply_patch(src, patch)
+        self.assertEqual(res, dst)
+
+    def test_minimal_patch(self):
+        """ Test whether a minimal patch is created, see #36 """
+        src = [{"foo": 1, "bar": 2}]
+        dst = [{"foo": 2, "bar": 2}]
+
+        patch = jsonpatch.make_patch(src, dst)
+
+        exp = [
+            {
+                "path": "/0/foo",
+                "value": 2,
+                "op": "replace"
+            }
+        ]
+
+        self.assertEqual(patch.patch, exp)
+
+
+class ListTests(unittest.TestCase):
+
+    def test_fail_prone_list_1(self):
+        """ Test making and applying a patch of the root is a list """
+        src = ['a', 'r', 'b']
+        dst = ['b', 'o']
+        patch = jsonpatch.make_patch(src, dst)
+        res = patch.apply(src)
+        self.assertEqual(res, dst)
+
+    def test_fail_prone_list_2(self):
+        """ Test making and applying a patch of the root is a list """
+        src = ['a', 'r', 'b', 'x', 'm', 'n']
+        dst = ['b', 'o', 'm', 'n']
+        patch = jsonpatch.make_patch(src, dst)
+        res = patch.apply(src)
+        self.assertEqual(res, dst)
+
+    def test_fail_prone_list_3(self):
+        """ Test making and applying a patch of the root is a list """
+        src = ['boo1', 'bar', 'foo1', 'qux']
+        dst = ['qux', 'bar']
+        patch = jsonpatch.make_patch(src, dst)
+        res = patch.apply(src)
+        self.assertEqual(res, dst)
+
+    def test_fail_prone_list_4(self):
+        """ Test making and applying a patch of the root is a list """
+        src = ['bar1', 59, 'foo1', 'foo']
+        dst = ['foo', 'bar', 'foo1']
+        patch = jsonpatch.make_patch(src, dst)
+        res = patch.apply(src)
+        self.assertEqual(res, dst)
+
 
 class InvalidInputTests(unittest.TestCase):
 
@@ -414,30 +608,54 @@ class ConflictTests(unittest.TestCase):
         self.assertRaises(jsonpatch.JsonPatchConflict, jsonpatch.apply_patch, src, patch_obj)
 
 
+class JsonPointerTests(unittest.TestCase):
 
-modules = ['jsonpatch']
+    def test_create_with_pointer(self):
+
+        patch = jsonpatch.JsonPatch([
+            {'op': 'add', 'path': jsonpointer.JsonPointer('/foo'), 'value': 'bar'},
+            {'op': 'add', 'path': jsonpointer.JsonPointer('/baz'), 'value': [1, 2, 3]},
+            {'op': 'remove', 'path': jsonpointer.JsonPointer('/baz/1')},
+            {'op': 'test', 'path': jsonpointer.JsonPointer('/baz'), 'value': [1, 3]},
+            {'op': 'replace', 'path': jsonpointer.JsonPointer('/baz/0'), 'value': 42},
+            {'op': 'remove', 'path': jsonpointer.JsonPointer('/baz/1')},
+            {'op': 'move', 'from': jsonpointer.JsonPointer('/foo'), 'path': jsonpointer.JsonPointer('/bar')},
+
+        ])
+        doc = {}
+        result = patch.apply(doc)
+        expected = {'bar': 'bar', 'baz': [42]}
+        self.assertEqual(result, expected)
 
 
-def get_suite():
-    suite = unittest.TestSuite()
-    suite.addTest(doctest.DocTestSuite(jsonpatch))
-    suite.addTest(unittest.makeSuite(ApplyPatchTestCase))
-    suite.addTest(unittest.makeSuite(EqualityTestCase))
-    suite.addTest(unittest.makeSuite(MakePatchTestCase))
-    suite.addTest(unittest.makeSuite(InvalidInputTests))
-    suite.addTest(unittest.makeSuite(ConflictTests))
-    return suite
+
+if __name__ == '__main__':
+    modules = ['jsonpatch']
 
 
-suite = get_suite()
+    def get_suite():
+        suite = unittest.TestSuite()
+        suite.addTest(doctest.DocTestSuite(jsonpatch))
+        suite.addTest(unittest.makeSuite(ApplyPatchTestCase))
+        suite.addTest(unittest.makeSuite(EqualityTestCase))
+        suite.addTest(unittest.makeSuite(MakePatchTestCase))
+        suite.addTest(unittest.makeSuite(ListTests))
+        suite.addTest(unittest.makeSuite(InvalidInputTests))
+        suite.addTest(unittest.makeSuite(ConflictTests))
+        suite.addTest(unittest.makeSuite(OptimizationTests))
+        suite.addTest(unittest.makeSuite(JsonPointerTests))
+        return suite
 
-for module in modules:
-    m = __import__(module, fromlist=[module])
-    suite.addTest(doctest.DocTestSuite(m))
 
-runner = unittest.TextTestRunner(verbosity=1)
+    suite = get_suite()
 
-result = runner.run(suite)
+    for module in modules:
+        m = __import__(module, fromlist=[module])
+        suite.addTest(doctest.DocTestSuite(m))
 
-if not result.wasSuccessful():
-    sys.exit(1)
+    runner = unittest.TextTestRunner(verbosity=1)
+
+    result = runner.run(suite)
+
+    if not result.wasSuccessful():
+        sys.exit(1)
